import operator

from karabo.bound import (
    DOUBLE_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
)
import numpy as np
from .base_kernel import BaseArbiterKernel, KernelWarning


class ReduceAndThreshold(BaseArbiterKernel):
    def __init__(self, config):
        super().__init__(config)
        self._key = config.get("key")
        self._threshold = config.get("threshold")
        self._reduction = getattr(np, config.get("reduction"))
        self._comparator = getattr(operator, config.get("comparator"))

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            NODE_ELEMENT(schema)
            .key(prefix)
            .description(
                "This kernel allows custom frame masks based on comparing some "
                "reduction along modules over each frame to some threshold. "
                "Integrated intensity will be used as a running example in the "
                "descriptions of the properties under this node; these settings (used "
                "with the example integrated intensity addon) would yield a frame "
                "mask of frames with total (across modules) integrated intensity of "
                "at least [threshold] ADUs."
            )
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.key")
            .description(
                "The key in the input hashes which holds the data values for reduction "
                "and comparison. Typically some summary computed per frame by an "
                "addon; for example, 'image.integratedIntensity'"
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.threshold")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.reduction")
            .description(
                "The reduction to apply across modules before comparing with the "
                "threshold. For example 'sum'."
            )
            .assignmentOptional()
            .defaultValue("sum")
            .options("sum,mean,max,min,std")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.comparator")
            .description(
                "The comparator to use for comparing the reduced per-frame values to "
                "the threshold. For the integrated intensity example, 'ge'. The "
                "reduced values will be on the left and the threshold will be on the "
                "right, so the example would be compute: "
                "mask = sum(integrated intensity) ≥ threshold."
            )
            .assignmentOptional()
            .defaultValue("ge")
            .options("lt,le,gt,ge,eq,ne")
            .reconfigurable()
            .commit(),
        )

    def consider(self, train_id, sources, num_frames):
        with self.warning_context(KernelWarning.MISSINGKEY) as warn:
            sources_with_key = [
                data[self._key] for (data, _) in sources.values() if data.has(self._key)
            ]
            if not sources_with_key:
                warn(f"No sources had '{self._key}'")
            else:
                return self._comparator(
                    self._reduction(sources_with_key, axis=0), self._threshold
                ).astype(np.uint8, copy=False)
        return np.ones(num_frames, dtype=np.uint8)
