import enum

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from .._version import version as deviceVersion


class Constants(enum.Enum):
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesPC = enum.auto()
    SlopesFF = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


bad_pixel_constants = {
    Constants.BadPixelsDark,
    Constants.BadPixelsPC,
    Constants.BadPixelsFF,
}


# from pycalibration's enum.py
class GainModes(enum.IntEnum):
    ADAPTIVE_GAIN = 0
    FIXED_HIGH_GAIN = 1
    FIXED_MEDIUM_GAIN = 2
    FIXED_LOW_GAIN = 3


# note: if this is extended further, bump up the dtype passed to kernel
class CorrectionFlags(enum.IntFlag):
    NONE = 0
    THRESHOLD = 1
    OFFSET = 2
    BLSHIFT = 4
    REL_GAIN_PC = 8
    GAIN_XRAY = 16
    BPMASK = 32
    FORCE_MG_IF_BELOW = 64
    FORCE_HG_IF_BELOW = 128


class AgipdBaseRunner(base_kernel_runner.BaseKernelRunner):
    _corrected_axis_order = "fxy"

    @property
    def input_shape(self):
        return (self.frames, 2, self.pixels_x, self.pixels_y)

    @property
    def processed_shape(self):
        return (self.frames, self.pixels_x, self.pixels_y)

    @property
    def map_shape(self):
        return (self.constant_memory_cells, self.pixels_x, self.pixels_y)

    @property
    def gm_map_shape(self):
        return self.map_shape + (3,)  # for gain-mapped constants

    @property
    def threshold_map_shape(self):
        return self.map_shape + (2,)

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
        bad_pixel_mask_value=np.nan,
        gain_mode=GainModes.ADAPTIVE_GAIN,
        g_gain_value=1,
        mg_hard_threshold=2000,
        hg_hard_threshold=2000,
        override_md_additional_offset=None,
    ):
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )
        self.gain_mode = gain_mode
        if self.gain_mode is GainModes.ADAPTIVE_GAIN:
            self.default_gain = self._xp.uint8(gain_mode)
        else:
            self.default_gain = self._xp.uint8(gain_mode - 1)
        self.gain_map = self._xp.empty(self.processed_shape, dtype=np.float32)

        # constants
        self.gain_thresholds = self._xp.empty(
            self.threshold_map_shape, dtype=np.float32
        )
        self.offset_map = self._xp.empty(self.gm_map_shape, dtype=np.float32)
        self.rel_gain_pc_map = self._xp.empty(self.gm_map_shape, dtype=np.float32)
        # not gm_map_shape because it only applies to medium gain pixels
        self.md_additional_offset = self._xp.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_xray_map = self._xp.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = self._xp.empty(self.gm_map_shape, dtype=np.uint32)
        self.override_md_additional_offset(override_md_additional_offset)
        self.set_bad_pixel_mask_value(bad_pixel_mask_value)
        self.set_mg_hard_threshold(mg_hard_threshold)
        self.set_hg_hard_threshold(hg_hard_threshold)
        self.set_g_gain_value(g_gain_value)
        self.flush_buffers(set(Constants))
        self.processed_data = self._xp.empty(self.processed_shape, output_data_dtype)

    @property
    def preview_data_views(self):
        return (
            self.input_data[:, 0],  # raw
            self.processed_data,  # corrected
            self.input_data[:, 1],  # raw gain
            self.gain_map,  # digitized gain
        )

    def load_constant(self, constant, constant_data):
        if constant is Constants.ThresholdsDark:
            # shape: y, x, memory cell, thresholds and gain values
            # note: the gain values are something like means used to derive thresholds
            self.gain_thresholds[:] = self._xp.asarray(
                constant_data[..., :2], dtype=np.float32
            ).transpose((2, 1, 0, 3))[:self.constant_memory_cells]
        elif constant is Constants.Offset:
            # shape: y, x, memory cell, gain stage
            self.offset_map[:] = self._xp.asarray(
                constant_data, dtype=np.float32
            ).transpose((2, 1, 0, 3))[:self.constant_memory_cells]
        elif constant is Constants.SlopesPC:
            # pc has funny shape (11, 352, 128, 512) from file
            # this is (fi, memory cell, y, x)
            # the following may contain NaNs, though...
            hg_slope = constant_data[0]
            hg_intercept = constant_data[1]
            mg_slope = constant_data[3]
            mg_intercept = constant_data[4]
            # TODO: remove sanitization (should happen in constant preparation notebook)
            # from agipdlib.py: replace NaN with median (per memory cell)
            # note: suffixes in agipdlib are "_m" and "_l", should probably be "_I"
            for naughty_array in (hg_slope, hg_intercept, mg_slope, mg_intercept):
                medians = np.nanmedian(naughty_array, axis=(1, 2))
                nan_bool = np.isnan(naughty_array)
                nan_cell, _, _ = np.where(nan_bool)
                naughty_array[nan_bool] = medians[nan_cell]

                too_low_bool = naughty_array < 0.8 * medians[:, np.newaxis, np.newaxis]
                too_low_cell, _, _ = np.where(too_low_bool)
                naughty_array[too_low_bool] = medians[too_low_cell]

                too_high_bool = naughty_array > 1.2 * medians[:, np.newaxis, np.newaxis]
                too_high_cell, _, _ = np.where(too_high_bool)
                naughty_array[too_high_bool] = medians[too_high_cell]

            frac_hg_mg = hg_slope / mg_slope
            rel_gain_map = np.ones(
                (3, self.constant_memory_cells, self.pixels_y, self.pixels_x),
                dtype=np.float32,
            )
            rel_gain_map[1] = rel_gain_map[0] * frac_hg_mg
            rel_gain_map[2] = rel_gain_map[1] * 4.48
            self.rel_gain_pc_map[:] = self._xp.asarray(
                rel_gain_map.transpose((1, 3, 2, 0)), dtype=np.float32
            )[:self.constant_memory_cells]
            if self._md_additional_offset_value is None:
                md_additional_offset = (
                    hg_intercept - mg_intercept * frac_hg_mg
                ).astype(np.float32)
                self.md_additional_offset[:] = self._xp.asarray(
                    md_additional_offset.transpose((0, 2, 1)), dtype=np.float32
                )[:self.constant_memory_cells]
        elif constant is Constants.SlopesFF:
            # constant shape: y, x, memory cell
            if constant_data.shape[2] == 2:
                # TODO: remove support for old format
                # old format, is per pixel only (shape is y, x, 2)
                # note: we should not support this in online
                constant_data = self._xp.broadcast_to(
                    constant_data[..., 0][..., np.newaxis],
                    (self.pixels_y, self.pixels_x, self.constant_memory_cells),
                )
            self.rel_gain_xray_map[:] = self._xp.asarray(
                constant_data.transpose(), dtype=np.float32
            )[:self.constant_memory_cells]
        else:
            assert constant in bad_pixel_constants
            # will simply OR with already loaded, does not take into account which ones
            constant_data = self._xp.asarray(constant_data, dtype=np.uint32)
            if len(constant_data.shape) == 3:
                if constant_data.shape == (
                    self.pixels_y,
                    self.pixels_x,
                    self.constant_memory_cells,
                ):
                    # BadPixelsFF is not per gain stage - broadcasting along gain
                    constant_data = self._xp.broadcast_to(
                        constant_data.transpose()[..., np.newaxis],
                        self.gm_map_shape,
                    )
                elif constant_data.shape == (
                    self.constant_memory_cells,
                    self.pixels_y,
                    self.pixels_x,
                ):
                    # old BadPixelsPC have different axis order
                    constant_data = self._xp.broadcast_to(
                        constant_data.transpose((0, 2, 1))[..., np.newaxis],
                        self.gm_map_shape,
                    )
                else:
                    raise ValueError(
                        f"Unsupported bad pixel map shape: {constant_data.shape}"
                    )
            else:
                # gain mapped constants seem consistent
                constant_data = constant_data.transpose((2, 1, 0, 3))
            self.bad_pixel_map |= constant_data[:self.constant_memory_cells]

    def override_md_additional_offset(self, override_value):
        self._md_additional_offset_value = override_value
        if override_value is not None:
            self.md_additional_offset.fill(override_value)

    def set_g_gain_value(self, override_value):
        self.g_gain_value = self._xp.float32(override_value)

    def set_bad_pixel_mask_value(self, mask_value):
        self.bad_pixel_mask_value = self._xp.float32(mask_value)

    def set_mg_hard_threshold(self, value):
        self.mg_hard_threshold = self._xp.float32(value)

    def set_hg_hard_threshold(self, value):
        self.hg_hard_threshold = self._xp.float32(value)

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)
        if Constants.SlopesPC in constants:
            self.rel_gain_pc_map.fill(1)
            self.md_additional_offset.fill(0)
        if Constants.SlopesFF:
            self.rel_gain_xray_map.fill(1)
        if constants & bad_pixel_constants:
            self.bad_pixel_map.fill(0)
            self.bad_pixel_map[
                :, 64:512:64
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value
            self.bad_pixel_map[
                :, 63:511:64
            ] |= utils.BadPixelValues.NON_STANDARD_SIZE.value


class AgipdCpuRunner(AgipdBaseRunner):
    _xp = np
    _gpu_based = False

    def correct(self, flags):
        self.correction_kernel(
            self.input_data,
            self.cell_table,
            flags,
            self.default_gain,
            self.gain_thresholds,
            self.mg_hard_threshold,
            self.hg_hard_threshold,
            self.offset_map,
            self.rel_gain_pc_map,
            self.md_additional_offset,
            self.rel_gain_xray_map,
            self.g_gain_value,
            self.bad_pixel_map,
            self.bad_pixel_mask_value,
            self.processed_data,
        )

    def _post_init(self):
        self.input_data = None
        self.cell_table = None
        # NOTE: CPU kernel does not yet support anything other than float32
        self.processed_data = self._xp.empty(
            self.processed_shape, dtype=self.output_data_dtype
        )
        from ..kernels import agipd_cython
        self.correction_kernel = agipd_cython.correct

    def load_data(self, image_data, cell_table):
        self.input_data = image_data
        self.cell_table = cell_table


class AgipdGpuRunner(AgipdBaseRunner):
    _gpu_based = True

    def _pre_init(self):
        import cupy as cp

        self._xp = cp

    def _post_init(self):
        self.input_data = self._xp.empty(self.input_shape, dtype=np.uint16)
        self.cell_table = self._xp.empty(self.frames, dtype=np.uint16)
        self.block_shape = (1, 1, 64)
        self.grid_shape = utils.grid_to_cover_shape_with_blocks(
            self.processed_shape, self.block_shape
        )
        self.correction_kernel = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("agipd_gpu.cu").render(
                {
                    "pixels_x": self.pixels_x,
                    "pixels_y": self.pixels_y,
                    "frames": self.frames,
                    "constant_memory_cells": self.constant_memory_cells,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self.output_data_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                }
            )
        ).get_function("correct")

    def load_data(self, image_data, cell_table):
        self.input_data.set(image_data)
        self.cell_table.set(cell_table)

    def correct(self, flags):
        if flags & CorrectionFlags.BLSHIFT:
            raise NotImplementedError("Baseline shift not implemented yet")

        self.correction_kernel(
            self.grid_shape,
            self.block_shape,
            (
                self.input_data,
                self.cell_table,
                self._xp.uint8(flags),
                self.default_gain,
                self.gain_thresholds,
                self.mg_hard_threshold,
                self.hg_hard_threshold,
                self.offset_map,
                self.rel_gain_pc_map,
                self.md_additional_offset,
                self.rel_gain_xray_map,
                self.g_gain_value,
                self.bad_pixel_map,
                self.bad_pixel_mask_value,
                self.gain_map,
                self.processed_data,
            ),
        )


class AgipdCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.ThresholdsDark: self.dark_condition,
            Constants.Offset: self.dark_condition,
            Constants.SlopesPC: self.dark_condition,
            Constants.SlopesFF: self.illuminated_condition,
            Constants.BadPixelsDark: self.dark_condition,
            Constants.BadPixelsPC: self.dark_condition,
            Constants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(schema, "AGIPD-Type")

        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(352)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(300)
            .commit()
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.acquisitionRate")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.gainSetting")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.photonEnergy")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.gainMode")
            .tags("managed")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options(",".join(gain_mode.name for gain_mode in GainModes))
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Acquisition rate"] = self._get_param("acquisitionRate")

        # TODO: remove this workaround after CalCat update
        integration_time = self._get_param("integrationTime")
        if integration_time != 12:
            res["Integration Time"] = integration_time

        gain_mode = GainModes[self._get_param("gainMode")]
        if gain_mode is not GainModes.ADAPTIVE_GAIN:
            res["Gain Mode"] = 1

        # TODO: make configurable whether or not to include gain setting?
        res["Gain Setting"] = self._get_param("gainSetting")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()

        # note: can consider always setting memory cells to 352 for FF
        # (deviation on constants in database should remove need for this, though)

        # for now, FF constants don't care about gain mode
        if "Gain Mode" in res:
            del res["Gain Mode"]

        res["Source Energy"] = self._get_param("photonEnergy")

        return res


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(base_correction.BaseCorrection):
    # subclass *must* set these attributes
    _base_output_schema = schemas.xtdf_output_schema()
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        # step name (used in schema), flag to enable for kernel, constants required
        ("thresholding", CorrectionFlags.THRESHOLD, {Constants.ThresholdsDark}),
        ("offset", CorrectionFlags.OFFSET, {Constants.Offset}),
        (  # TODO: specify that /both/ constants are needed, not just one or the other
            "forceMgIfBelow",
            CorrectionFlags.FORCE_MG_IF_BELOW,
            {Constants.ThresholdsDark, Constants.Offset},
        ),
        (
            "forceHgIfBelow",
            CorrectionFlags.FORCE_HG_IF_BELOW,
            {Constants.ThresholdsDark, Constants.Offset},
        ),
        ("relGainPc", CorrectionFlags.REL_GAIN_PC, {Constants.SlopesPC}),
        ("gainXray", CorrectionFlags.GAIN_XRAY, {Constants.SlopesFF}),
        (
            "badPixels",
            CorrectionFlags.BPMASK,
            bad_pixel_constants | {
                None,  # means stay available even without constants loaded
            },
        ),
    )
    _calcat_friend_class = AgipdCalcatFriend
    _constant_enum_class = Constants
    _preview_outputs = [
        "outputRaw",
        "outputCorrected",
        "outputRawGain",
        "outputGainMap",
    ]

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(AgipdCorrection._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(352)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("cell")
            .commit(),
        )

        # this is not automatically done by superclass for complicated class reasons
        AgipdCalcatFriend.add_schema(expected)
        base_correction.add_addon_nodes(expected, AgipdCorrection)
        base_correction.add_correction_step_schema(
            expected,
            AgipdCorrection._correction_steps,
        )
        base_correction.add_bad_pixel_config_node(expected)
        base_correction.add_preview_outputs(expected, AgipdCorrection._preview_outputs)
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

        # turn off the force MG / HG steps by default
        for step in ("forceMgIfBelow", "forceHgIfBelow"):
            for flag in ("enable", "preview"):
                (
                    OVERWRITE_ELEMENT(expected)
                    .key(f"corrections.{step}.{flag}")
                    .setNewDefaultValue(False)
                    .commit(),
                )
        # additional settings specific to AGIPD correction steps
        (
            FLOAT_ELEMENT(expected)
            .key("corrections.forceMgIfBelow.hardThreshold")
            .tags("managed")
            .description(
                "If enabled, any pixels assigned to low gain stage which would be "
                "below this threshold after having medium gain offset subtracted will "
                "be reassigned to medium gain."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.forceHgIfBelow.hardThreshold")
            .tags("managed")
            .description(
                "Like forceMgIfBelow, but potentially reassigning from medium gain to "
                "high gain based on threshold and pixel value minus low gain offset. "
                "Applied after forceMgIfBelow, pixels can theoretically be reassigned "
                "twice, from LG to MG and to HG."
            )
            .assignmentOptional()
            .defaultValue(1000)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("corrections.relGainPc.overrideMdAdditionalOffset")
            .tags("managed")
            .displayedName("Override md_additional_offset")
            .description(
                "Toggling this on will use the value in the next field globally for "
                "md_additional_offset. Note that the correction map on GPU gets "
                "overwritten as long as this boolean is True, so reload constants "
                "after turning off."
            )
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.relGainPc.mdAdditionalOffset")
            .tags("managed")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected)
            .key("corrections.gainXray.gGainValue")
            .tags("managed")
            .displayedName("G_gain_value")
            .description(
                "Newer X-ray gain correction constants are absolute. The default "
                "G_gain_value of 1 means that output is expected to be in keV. If "
                "this is not desired, one can here specify the mean X-ray gain value "
                "over all modules to get ADU values out - operator must manually "
                "find this mean value."
            )
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),
        )

    @property
    def input_data_shape(self):
        return (
            self.unsafe_get("dataFormat.frames"),
            2,
            self.unsafe_get("dataFormat.pixelsX"),
            self.unsafe_get("dataFormat.pixelsY"),
        )

    def __init__(self, config):
        super().__init__(config)
        # note: gain mode single sourced from constant retrieval node
        try:
            np.float32(config.get("corrections.badPixels.maskingValue"))
        except ValueError:
            config["corrections.badPixels.maskingValue"] = "nan"

        self._has_updated_bad_pixel_selection = False

    @property
    def bad_pixel_mask_value(self):
        return np.float32(self.unsafe_get("corrections.badPixels.maskingValue"))

    _override_bad_pixel_flags = property(base_correction.get_bad_pixel_field_selection)

    @property
    def _kernel_runner_init_args(self):
        return {
            "gain_mode": self.gain_mode,
            "bad_pixel_mask_value": self.bad_pixel_mask_value,
            "g_gain_value": self.unsafe_get("corrections.gainXray.gGainValue"),
            "mg_hard_threshold": self.unsafe_get(
                "corrections.forceMgIfBelow.hardThreshold"
            ),
            "hg_hard_threshold": self.unsafe_get(
                "corrections.forceHgIfBelow.hardThreshold"
            ),
        }

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return AgipdCpuRunner
        else:
            return AgipdGpuRunner

    @property
    def gain_mode(self):
        return GainModes[self.unsafe_get("constantParameters.gainMode")]

    @property
    def _override_md_additional_offset(self):
        if self.unsafe_get("corrections.relGainPc.overrideMdAdditionalOffset"):
            return self.unsafe_get("corrections.relGainPc.mdAdditionalOffset")
        else:
            return None

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        pulse_table,
    ):
        """Called by input_handler for each data hash. Should correct data, optionally
        compute preview, write data output, and optionally write preview outputs."""
        # original shape: frame, data/raw_gain, x, y
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            self.kernel_runner.load_data(image_data, cell_table.ravel())
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            self.log_status_warn(
                f"For debugging - cell table: {type(cell_table)}, {cell_table}"
            )
            self.log_status_warn(
                f"For debugging - data: {type(image_data)}, {image_data.shape}"
            )
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        # first prepare previews (not affected by addons)
        self.kernel_runner.correct(self._correction_flag_preview)
        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
        (
            preview_raw,
            preview_corrected,
            preview_raw_gain,
            preview_gain_map,
        ) = self.kernel_runner.compute_previews(preview_slice_index)
        # TODO: start writing out previews asynchronously in the background

        # then prepare full corrected data
        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        if self._correction_flag_enabled != self._correction_flag_preview:
            self.kernel_runner.correct(self._correction_flag_enabled)
        for addon in self._enabled_addons:
            addon.post_correction(
                self.kernel_runner.processed_data, cell_table, pulse_table, data_hash
            )
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        for addon in self._enabled_addons:
            addon.post_reshape(
                buffer_array, cell_table, pulse_table, data_hash
            )

        # reusing input data hash for sending
        if self._use_shmem_handles:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
        data_hash.set(self._pulse_table_path, pulse_table[:, np.newaxis])

        self._write_output(data_hash, metadata)
        self._preview_friend.write_outputs(
            metadata, preview_raw, preview_corrected, preview_raw_gain, preview_gain_map
        )

    def _load_constant_to_runner(self, constant, constant_data):
        if constant in bad_pixel_constants:
            constant_data &= self._override_bad_pixel_flags
        self.kernel_runner.load_constant(constant, constant_data)

    def preReconfigure(self, config):
        super().preReconfigure(config)
        if config.has("corrections.badPixels.maskingValue"):
            # only check if it is valid; postReconfigure will use it
            np.float32(config.get("corrections.badPixels.maskingValue"))

    def postReconfigure(self):
        super().postReconfigure()

        update = self._prereconfigure_update_hash

        if update.has("constantParameters.gainMode"):
            self.flush_constants()
            self._update_buffers()

        if update.has("corrections.forceMgIfBelow.hardThreshold"):
            self.kernel_runner.set_mg_hard_threshold(
                self.get("corrections.forceMgIfBelow.hardThreshold")
            )

        if update.has("corrections.forceHgIfBelow.hardThreshold"):
            self.kernel_runner.set_hg_hard_threshold(
                self.get("corrections.forceHgIfBelow.hardThreshold")
            )

        if update.has("corrections.gainXray.gGainValue"):
            self.kernel_runner.set_g_gain_value(
                self.get("corrections.gainXray.gGainValue")
            )

        if update.has("corrections.badPixels.maskingValue"):
            self.kernel_runner.set_bad_pixel_mask_value(self.bad_pixel_mask_value)

        if update.has("corrections.badPixels.subsetToUse"):
            self.log_status_info("Updating bad pixel maps based on subset specified")
            # note: now just always reloading from cache for convenience
            with self.calcat_friend.cached_constants_lock:
                self.kernel_runner.flush_buffers(bad_pixel_constants)
                for constant in bad_pixel_constants:
                    if constant in self.calcat_friend.cached_constants:
                        self._load_constant_to_runner(
                            constant, self.calcat_friend.cached_constants[constant]
                        )
