import concurrent.futures
import enum
import functools

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
)

from .. import (
    base_calcat,
    base_correction,
    base_kernel_runner,
    schemas,
    utils,
)
from .._version import version as deviceVersion
from ..kernels import common_cython


class Constants(enum.Enum):
    BadPixelsDarkEPix100 = enum.auto()
    RelativeGainEPix100 = enum.auto()
    OffsetEPix100 = enum.auto()
    NoiseEPix100 = enum.auto()


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 2**1
    COMMONMODE = 2**2
    RELGAIN = 2**3
    BPMASK = 2**4


class Epix100CalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {
            Constants.OffsetEPix100: self.dark_condition,
            Constants.BadPixelsDarkEPix100: self.dark_condition,
            Constants.NoiseEPix100: self.dark_condition,
            Constants.RelativeGainEPix100: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(schema):
        super(Epix100CalcatFriend, Epix100CalcatFriend).add_schema(
            schema, "ePix100-Type"
        )

        # set some defaults for common parameters
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(200)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .setNewDefaultValue(708)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .setNewDefaultValue(768)
            .commit(),
        )

        (
            DOUBLE_ELEMENT(schema)
            .key("constantParameters.integrationTime")
            .tags("managed")
            .displayedName("Integration Time")
            .assignmentOptional()
            .defaultValue(10)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sensorTemperature")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(264.32933824373004)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.inVacuum")
            .tags("managed")
            .displayedName("In vacuum")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.sourceEnergy")
            .tags("managed")
            .displayedName("Source Energy")
            .assignmentOptional()
            .defaultValue(8.04778)
            .reconfigurable()
            .commit(),
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Integration Time"] = self._get_param("integrationTime")
        res["Sensor Temperature"] = self._get_param("sensorTemperature")
        res["In vacuum"] = self._get_param("inVacuum")

        return res

    def illuminated_condition(self):
        res = self.dark_condition()

        res["Source Energy"] = self._get_param("sourceEnergy")

        return res


class Epix100CpuRunner(base_kernel_runner.BaseKernelRunner):
    _corrected_axis_order = "xy"
    _xp = np
    _gpu_based = False

    @property
    def input_shape(self):
        return (self.pixels_x, self.pixels_y)

    @property
    def preview_shape(self):
        return (self.pixels_x, self.pixels_y)

    @property
    def processed_shape(self):
        return self.input_shape

    @property
    def map_shape(self):
        return (self.pixels_x, self.pixels_y)

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,  # will be 1, will be ignored
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        assert (
            output_data_dtype == np.float32
        ), "Alternative output types not supported yet"

        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )

        self.input_data = None
        self.processed_data = np.empty(self.processed_shape, dtype=np.float32)

        self.offset_map = np.empty(self.map_shape, dtype=np.float32)
        self.rel_gain_map = np.empty(self.map_shape, dtype=np.float32)
        self.bad_pixel_map = np.empty(self.map_shape, dtype=np.uint32)
        self.noise_map = np.empty(self.map_shape, dtype=np.float32)
        # will do everything by quadrant
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=4)
        self._q_input_data = None
        self._q_processed_data = utils.quadrant_views(self.processed_data)
        self._q_offset_map = utils.quadrant_views(self.offset_map)
        self._q_rel_gain_map = utils.quadrant_views(self.rel_gain_map)
        self._q_bad_pixel_map = utils.quadrant_views(self.bad_pixel_map)
        self._q_noise_map = utils.quadrant_views(self.noise_map)

    def __del__(self):
        self.thread_pool.shutdown()

    @property
    def preview_data_views(self):
        # NOTE: apparently there are "calibration rows"
        # geometry assumes these are cut out already
        return (self.input_data[2:-2], self.processed_data[2:-2])

    def load_data(self, image_data):
        # should almost never squeeze, but ePix doesn't do burst mode, right?
        self.input_data = image_data.astype(np.uint16, copy=False).squeeze()
        self._q_input_data = utils.quadrant_views(self.input_data)

    def load_constant(self, constant_type, data):
        if constant_type is Constants.OffsetEPix100:
            self.offset_map[:] = data.squeeze().astype(np.float32)
        elif constant_type is Constants.RelativeGainEPix100:
            self.rel_gain_map[:] = data.squeeze().astype(np.float32)
        elif constant_type is Constants.BadPixelsDarkEPix100:
            self.bad_pixel_map[:] = data.squeeze()
        elif constant_type is Constants.NoiseEPix100:
            self.noise_map[:] = data.squeeze()
        else:
            raise Exception(f"Unknown constant type {constant_type}")

    def flush_buffers(self, constants):
        if Constants.OffsetEPix100 in constants:
            self.offset_map.fill(0)
        if Constants.RelativeGainEPix100 in constants:
            self.rel_gain_map.fill(1)
        if constants & {Constants.BadPixelsDarkEPix100}:
            self.bad_pixel_map.fill(0)
        if Constants.NoiseEPix100 in constants:
            self.noise_map.fill(np.inf)

    def _correct_quadrant(
        self,
        q,
        flags,
        bad_pixel_mask_value,
        cm_noise_sigma,
        cm_min_frac,
        cm_row,
        cm_col,
        cm_block,
    ):
        output = self._q_processed_data[q]
        output[:] = self._q_input_data[q].astype(np.float32)
        if flags & CorrectionFlags.OFFSET:
            output -= self._q_offset_map[q]

        if flags & CorrectionFlags.COMMONMODE:
            # per rectangular block that looks like something is going on
            cm_mask = (
                (self._q_bad_pixel_map[q] != 0)
                | (output > self._q_noise_map[q] * cm_noise_sigma)
            ).astype(np.uint8, copy=False)
            masked = np.ma.masked_array(data=output, mask=cm_mask)
            if cm_block:
                for block in np.hsplit(masked, 4):
                    if block.count() < block.size * cm_min_frac:
                        continue
                    block.data[:] -= np.ma.median(block)

            if cm_row:
                common_cython.cm_fs(output, cm_mask, cm_noise_sigma, cm_min_frac)

            if cm_col:
                common_cython.cm_ss(output, cm_mask, cm_noise_sigma, cm_min_frac)

        if flags & CorrectionFlags.RELGAIN:
            output *= self._q_rel_gain_map[q]

        if flags & CorrectionFlags.BPMASK:
            output[self._q_bad_pixel_map[q] != 0] = bad_pixel_mask_value

    def correct(
        self,
        flags,
        bad_pixel_mask_value=np.nan,
        cm_noise_sigma=5,
        cm_min_frac=0.25,
        cm_row=True,
        cm_col=True,
        cm_block=True,
    ):
        # NOTE: how to best clean up all these duplicated parameters?
        for result in self.thread_pool.map(
            functools.partial(
                self._correct_quadrant,
                flags=flags,
                bad_pixel_mask_value=bad_pixel_mask_value,
                cm_noise_sigma=cm_noise_sigma,
                cm_min_frac=cm_min_frac,
                cm_row=cm_row,
                cm_col=cm_col,
                cm_block=cm_block,
            ),
            range(4),
        ):
            pass  # just run through to await map


@KARABO_CLASSINFO("Epix100Correction", deviceVersion)
class Epix100Correction(base_correction.BaseCorrection):
    _base_output_schema = schemas.jf_output_schema()
    _correction_flag_class = CorrectionFlags
    _correction_steps = (
        ("offset", CorrectionFlags.OFFSET, {Constants.OffsetEPix100}),
        ("relGain", CorrectionFlags.RELGAIN, {Constants.RelativeGainEPix100}),
        ("commonMode", CorrectionFlags.COMMONMODE, {Constants.NoiseEPix100}),
        ("badPixels", CorrectionFlags.BPMASK, {Constants.BadPixelsDarkEPix100}),
    )
    _image_data_path = "data.image.pixels"
    _kernel_runner_class = Epix100CpuRunner
    _calcat_friend_class = Epix100CalcatFriend
    _constant_enum_class = Constants
    _preview_outputs = ["outputRaw", "outputCorrected"]
    _cell_table_path = None
    _pulse_table_path = None
    _warn_memory_cell_range = False

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(Epix100Correction._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .setNewDefaultValue(708)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .setNewDefaultValue(768)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(1)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .setNewOptions("xy,yx")
            .setNewDefaultValue("xy")
            .commit(),

            # TODO: disable preview selection mode

            OVERWRITE_ELEMENT(expected)
            .key("useShmemHandles")
            .setNewDefaultValue(False)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .setNewDefaultValue(2)
            .commit(),
        )

        base_correction.add_preview_outputs(
            expected, Epix100Correction._preview_outputs
        )
        base_correction.add_correction_step_schema(
            expected,
            Epix100Correction._correction_steps,
        )
        (
            DOUBLE_ELEMENT(expected)
            .key("corrections.commonMode.noiseSigma")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(5)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("corrections.commonMode.minFrac")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(0.25)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("corrections.commonMode.enableRow")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("corrections.commonMode.enableCol")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("corrections.commonMode.enableBlock")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )
        Epix100CalcatFriend.add_schema(expected)
        # TODO: bad pixel node?

    @property
    def input_data_shape(self):
        # TODO: check
        return (
            self.unsafe_get("dataFormat.pixelsX"),
            self.unsafe_get("dataFormat.pixelsY"),
        )

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,  # will be None
        pulse_table,  # ditto
    ):
        self.kernel_runner.load_data(image_data)

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        args_which_should_be_cached = dict(
            cm_noise_sigma=self.unsafe_get("corrections.commonMode.noiseSigma"),
            cm_min_frac=self.unsafe_get("corrections.commonMode.minFrac"),
            cm_row=self.unsafe_get("corrections.commonMode.enableRow"),
            cm_col=self.unsafe_get("corrections.commonMode.enableCol"),
            cm_block=self.unsafe_get("corrections.commonMode.enableBlock"),
        )
        self.kernel_runner.correct(
            flags=self._correction_flag_enabled, **args_which_should_be_cached
        )
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        if self._correction_flag_enabled != self._correction_flag_preview:
            self.kernel_runner.correct(
                flags=self._correction_flag_preview,
                **args_which_should_be_cached,
            )

        if self._use_shmem_handles:
            # TODO: consider better name for key...
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        self._write_output(data_hash, metadata)

        # note: base class preview machinery assumes burst mode, shortcut it
        self._preview_friend.write_outputs(
            metadata, *self.kernel_runner.preview_data_views
        )

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_constant(constant, constant_data)
