import re

from karabo.bound import (
    BOOL_ELEMENT,
    NODE_ELEMENT,
    STRING_ELEMENT,
    VECTOR_STRING_ELEMENT,
    Hash,
)
import numpy as np


class FrameselectionFriend:
    @staticmethod
    def add_schema(schema):
        (
            NODE_ELEMENT(schema)
            .key("frameSelector")
            .displayedName("Frame selection")
            .commit(),

            BOOL_ELEMENT(schema)
            .key("frameSelector.enable")
            .displayedName("Enabled")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("frameSelector.arbiterSource")
            .displayedName("Arbiter source")
            .description(
                "Source name to pull the frame selection pattern from, must be part of "
                "matched sources."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("frameSelector.dataSourcePattern")
            .displayedName("Data source pattern")
            .description(
                "Source name pattern to apply frame selection to.  Should match "
                "subset of matched sources."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            VECTOR_STRING_ELEMENT(schema)
            .key("frameSelector.dataKeys")
            .displayedName("Data keys")
            .description("Keys in data sources to apply frame selection to.")
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),
        )

    def __init__(self, config):
        self._config = Hash()
        self._enabled = False
        self._arbiter = ""
        self._data_pattern = ""
        self._data_keys = []
        self.reconfigure(config)

    def reconfigure(self, config):
        self._config.merge(config)
        self._enabled = self._config.get("enable")
        self._arbiter = self._config.get("arbiterSource")
        self._source_pattern = re.compile(self._config.get("dataSourcePattern"))
        self._data_keys = list(self._config.get("dataKeys"))

    def get_mask(self, sources):
        if self._enabled and self._arbiter in sources:
            return np.array(
                sources[self._arbiter][0]["data.dataFramePattern"], copy=False
            ).astype(np.bool, copy=False)
        else:
            return None

    def apply_mask(self, source, data_hash, mask):
        if mask is not None and self._source_pattern.match(source):
            for key in self._data_keys:
                if not data_hash.has(key):
                    continue
                if data_hash[key].shape[0] != mask.size:
                    self._device.log.WARN(
                        "Frame selection mask does not match the data size"
                    )
                    continue

                data_hash[key] = data_hash[key][mask]
